//**************************************************************************//
//* FRIDA: fast reliable interactive data analysis                         *//
//* func.cpp: elementary functions for use in generic expressions          *//
//* (C) Joachim Wuttke 1990-, v2(C++) 2001-                                *//
//* http://frida.sourceforge.net                                           *//
//**************************************************************************//

#include <math.h>
#include <iostream>
#include <map>
#include <gsl/gsl_sf.h>

#include "mystd.h"
#include "../../code/kww/kww.h"
#include "rng.h"
#include "olm.h" // for func_data_, comes with many includes
#include "asi.h" // for find

double twopi = 8*atan(1.);

//**************************************************************************//
//*   Functions of one argument                                            *//
//**************************************************************************//

double func_neg(double v) { return -v; }
double func_inv(double v) { return v ? 1/v : 0; }

double func_ln(double v) { return v>0 ? log(v) : 0; }
double func_lg(double v) { return v>0 ? log10(v) : 0; }
	
double func_i_ln(double v) { return exp(v); }
double func_i_lg(double v) { return exp(log(10.)*v); }
	
double func_sqr(double v) { return v*v; }
double func_sqrt(double v) { return v>=0 ? sqrt(v) : 0; }

double func_abs(double v) { return v>=0 ? v : -v; }

double func_exp(double v) { return exp(v); }
	
double func_sin(double v) { return sin(v); }
double func_cos(double v) { return cos(v); }
double func_tan(double v) { return tan(v); }
double func_cot(double v) { double t=tan(v); return t==0?0:1/t;}

double func_sind(double v) { return sin(v*twopi/360.); }
double func_cosd(double v) { return cos(v*twopi/360.); }
double func_tand(double v) { return tan(v*twopi/360.); }
double func_cotd(double v) { double t=tan(v*twopi/360.);  return t==0?0:1/t; }
	
double func_asin(double v) { return (-1<=v && v<=1) ? asin(v) : 0; }
double func_acos(double v) { return (-1<=v && v<=1) ? acos(v) : 0; }
double func_atan(double v) { return atan(v); }
double func_acot(double v) { return v!=0 ? atan(1/v) : twopi/4; }

double func_asind(double v) { return (-1<=v && v<=1) ? 360./twopi*asin(v) : 0; }
double func_acosd(double v) { return (-1<=v && v<=1) ? 360./twopi*acos(v) : 0; }
double func_atand(double v) { return 360./twopi*atan(v); }
double func_acotd(double v) { return v!=0 ? 360./twopi*atan(1/v) : 90; }

double func_gamma(double v) { return v>0 ? gsl_sf_gamma(v) : 0; } 
// erweiterungsfhig (v<0 aber nichtganzzahlig ?)

double func_erfP(double v) { return 1-gsl_sf_erf_Q(v); }
double func_erfQ(double v) { return gsl_sf_erf_Q(v); } 
double func_erf(double v) { return gsl_sf_erf(v); } 
double func_erfc(double v) { return gsl_sf_erfc(v); } 

double func_sinc(double v) { return v==0 ? 1 : sin(v)/v; }

double func_ceil(double v)  { return ceil(v); }      // smallest int >= arg
double func_floor(double v) { return floor(v); }     // largest  int <= arg
double func_nint(double v)  { return round(v); } // nearest  int

//**************************************************************************//
//*   Functions of two arguments                                           *//
//**************************************************************************//

double func_add(double v, double a) { return v+a; }
double func_sub(double v, double a) { return v-a; }
double func_mul(double v, double a) { return v*a; }
double func_div(double v, double a) { return a!=0 ? v/a : 0; }
double func_idiv(double v, double a) { return a!=0 ? ((int)v)/((int)a) : 0; }
double func_pow(double v, double a) {
    if (v<0)
        return a!=(int)a ? 0 : pow(v,a);
    else if (v==0)
        return a<=0 ? 0 : pow(v,a);
    else
        return pow(v,a);
}

double func_2nd(double v, double a) { return a; }
double func_2div(double v, double a) { return v!=0 ? a/v : 0; }

double func_not(double v) { return !v; }

double func_eq (double v, double a) { return v == a; }
double func_ne (double v, double a) { return v != a; }
double func_gt (double v, double a) { return v > a; }
double func_ge (double v, double a) { return v >= a; }
double func_lt (double v, double a) { return v < a; }
double func_le (double v, double a) { return v <= a; }
double func_and(double v, double a) { return v && a; }
double func_or (double v, double a) { return v || a; }
double func_xor(double v, double a) { return (bool) v ^ (bool) a; }
double func_mod(double v, double a) { return fmod(v,a); }
double func_over(double v, double a) {
    return  ( 0<=v && 0<=a && a<=v && v <= GSL_SF_GAMMA_XMAX ) ?
        gsl_sf_gamma(v+1)/gsl_sf_gamma(a+1)/gsl_sf_gamma(v-a+1) :
        0; }

double func_min (double v, double a) { return v<a ? v : a; }
double func_max (double v, double a) { return v>a ? v : a; }
double func_gauss (double x, double s) {
    return s==0 ? 0 : exp(-x*x/(2*s*s))/sqrt(twopi)/s; }
double func_gaussnn (double x, double s) { // Gauss not normalized
    return s==0 ? 0 : exp(-x*x/(2*s*s)); }
double func_cauchy (double x, double s) {
    return s==0 ? 0 : 2*s/twopi/(x*x+s*s); }

double func_kww2( double v, double a){
    // printf( "DEB KWW2 %9.4g %9.4g ...\n", a, v );
    double h = kohlrausch_cos_spectrum( fabs(v), a<.1 ? .1 : a>2 ? 2 : a );
    // printf( "DEB KWW2 %9.4g %9.4g -> %9.4g\n", a, v, h );
    return h;
}
double func_kww1( double v, double a){
    return kohlrausch_sin_spectrum( v, a );
}

double func_ran (double v, double a) { 
    // a random number between v and a :
    return func_min(v,a) + (func_max(v,a)-func_min(v,a))*NRNG::uniform(); }

double func_in (double d1, vector<double> v2)
{
    for( uint i=0; i<v2.size(); ++i )
        if( d1==v2[i] ) return 1.;
    return 0.;
}

//**************************************************************************//
//*   Functions of three arguments                                         *//
//**************************************************************************//

double func_condass(double v, double a, double b) { return v ? a : b; }

double func_q4w( double hbarw, double Ei, double twotheta) { 
    // calculate Q in Angstroem-1 when hbar*w and E_i are given in meV
    if( hbarw>Ei ) return -1;
    return sqrt( (1/2.07212349) *
            ( 2*Ei-hbarw-2*sqrt(Ei*(Ei-hbarw))*cos(twotheta*twopi/360) ) );
}

double func_cauchy2 (double x, double p, double w) {
    return w==0 ? 0 : 2*w/twopi*(1.0/(SQR(x-p)+SQR(w))+1.0/(SQR(x+p)+SQR(w)));
}

//**************************************************************************//
//*   Functions based on data files                                        *//
//**************************************************************************//

//! linear interpolation in data from scan k,j.

double func_data_intp_lin( double v, double dk, double dj )
{
    uint k = lrint( dk );
    uint j = lrint( dj );
    if( k<0 || k>=NOlm::MOM.size() )
        throw string( "intp: invalid k" );
    const COld *fd = NOlm::MOM[k].D();
    if( !fd )
        throw "f" + strg(k) + " is no data file";
    if( j<0 || j>=fd->nScan() )
        throw string( "intp: invalid j" );
    const CScan *s = &(fd->VS[j]);
    // we assume that s->x is sorted
    uint ilow = 0, ihig = s->size()-1;
    if( v < s->x[ilow] || v > s->x[ihig] )
        return 0;
    while( ihig-ilow > 1 ){
        uint imid = (ilow+ihig)/2;
        if( v < s->x[imid] )
            ihig = imid;
        else
            ilow = imid;
    }
    return s->y[ilow] +
        (v-s->x[ilow])/(s->x[ihig]-s->x[ilow])*(s->y[ihig]-s->y[ilow]);
}

//**************************************************************************//
//*   Function registration and retrieval                                  *//
//**************************************************************************//

namespace NFunctions { // internals:
    map<string,CFunc> fmap;
};

void NFunctions::Register(void) {

// mon-op
    fmap.insert(make_pair(string("!"),    CFunc("!",    func_not)));

// bin-op  
    fmap.insert(make_pair(string("$"),    CFunc("$",    func_2nd)));
    fmap.insert(make_pair(string("+"),    CFunc("+",    func_add)));
    fmap.insert(make_pair(string("-"),    CFunc("-",    func_sub)));
    fmap.insert(make_pair(string("*"),    CFunc("*",    func_mul)));
    fmap.insert(make_pair(string("/"),    CFunc("/",    func_div)));
//    fmap.insert(make_pair(string("%"),    CFunc("&",    func_rmd)));
    fmap.insert(make_pair(string("$/"),   CFunc("$/",   func_2div)));
    fmap.insert(make_pair(string("^"),    CFunc("^",    func_pow)));
  
    fmap.insert(make_pair(string("=="),   CFunc("==",  func_eq)));
    fmap.insert(make_pair(string("!="),   CFunc("!=",  func_ne)));
    fmap.insert(make_pair(string(">"),    CFunc(">",   func_gt)));
    fmap.insert(make_pair(string(">="),   CFunc(">=",  func_ge)));
    fmap.insert(make_pair(string("<"),    CFunc("<",   func_lt)));
    fmap.insert(make_pair(string("<="),   CFunc("<=",  func_le)));
    fmap.insert(make_pair(string("&&"),   CFunc("&&",  func_and)));
    fmap.insert(make_pair(string("||"),   CFunc("||",  func_or)));
    fmap.insert(make_pair(string("xor"),  CFunc("xor", func_xor)));
    fmap.insert(make_pair(string("mod"),  CFunc("mod", func_mod)));
    fmap.insert(make_pair(string("idiv"), CFunc("mod", func_idiv)));
    fmap.insert(make_pair(string("over"), CFunc("over",func_over)));
    fmap.insert(make_pair(string("kww2"), CFunc("kww2",func_kww2)));
    fmap.insert(make_pair(string("kww1"), CFunc("kww1",func_kww1)));

// op-t-v
    fmap.insert(make_pair(string("in"),   CFunc("in",   func_in)));

// ter-op    
    fmap.insert(make_pair(string("?:"),   CFunc("?:",   func_condass)));

// f( 1 arg )
    fmap.insert(make_pair(string("0-"),   CFunc("0-",   func_neg)));
    fmap.insert(make_pair(string("1/"),   CFunc("1/",   func_inv)));
    fmap.insert(make_pair(string("ln"),   CFunc("ln",   func_ln)));
    fmap.insert(make_pair(string("lg"),   CFunc("lg",   func_lg)));
    fmap.insert(make_pair(string("e^"),   CFunc("e^",   func_i_ln)));
    fmap.insert(make_pair(string("10^"),  CFunc("10^",  func_i_lg)));
    fmap.insert(make_pair(string("^2"),   CFunc("^2",   func_sqr)));
    fmap.insert(make_pair(string("sqr"),  CFunc("sqr",  func_sqr)));
    fmap.insert(make_pair(string("sqrt"), CFunc("sqrt", func_sqrt)));
    fmap.insert(make_pair(string("abs"),  CFunc("abs",  func_abs)));
    fmap.insert(make_pair(string("exp"),  CFunc("exp",  func_exp)));
    fmap.insert(make_pair(string("sin"),  CFunc("sin",  func_sin)));
    fmap.insert(make_pair(string("cos"),  CFunc("cos",  func_cos)));
    fmap.insert(make_pair(string("tan"),  CFunc("tan",  func_tan)));
    fmap.insert(make_pair(string("cot"),  CFunc("cot",  func_cot)));
    fmap.insert(make_pair(string("sind"), CFunc("sind", func_sind)));
    fmap.insert(make_pair(string("cosd"), CFunc("cosd", func_cosd)));
    fmap.insert(make_pair(string("tand"), CFunc("tand", func_tand)));
    fmap.insert(make_pair(string("cotd"), CFunc("cotd", func_cotd)));
    fmap.insert(make_pair(string("asin"), CFunc("asin", func_asin)));
    fmap.insert(make_pair(string("acos"), CFunc("acos", func_acos)));
    fmap.insert(make_pair(string("atan"), CFunc("atan", func_atan)));
    fmap.insert(make_pair(string("acot"), CFunc("acot", func_acot)));
    fmap.insert(make_pair(string("asind"),CFunc("asind",func_asind)));
    fmap.insert(make_pair(string("acosd"),CFunc("acosd",func_acosd)));
    fmap.insert(make_pair(string("atand"),CFunc("atand",func_atand)));
    fmap.insert(make_pair(string("acotd"),CFunc("acotd",func_acotd)));
    fmap.insert(make_pair(string("gamma"),CFunc("gamma",func_gamma)));
    fmap.insert(make_pair(string("erfP"), CFunc("erfP", func_erfP)));
    fmap.insert(make_pair(string("erfQ"), CFunc("erfQ", func_erfQ)));
    fmap.insert(make_pair(string("erf"),  CFunc("erf",  func_erf)));
    fmap.insert(make_pair(string("erfc"), CFunc("erfc", func_erfc)));
    fmap.insert(make_pair(string("sinc"), CFunc("sinc", func_sinc)));
    fmap.insert(make_pair(string("ceil"), CFunc("ceil", func_ceil)));
    fmap.insert(make_pair(string("floor"),CFunc("floor",func_floor)));
    fmap.insert(make_pair(string("nint"), CFunc("nint", func_nint)));

// f ( 2arg )
    fmap.insert(make_pair(string("min"),  CFunc("min",  func_min)));
    fmap.insert(make_pair(string("max"),  CFunc("max",  func_max)));
    fmap.insert(make_pair(string("gauss"),CFunc("gauss",  func_gauss)));
    fmap.insert(make_pair(string("gaussnn"),CFunc("gaussnn",  func_gaussnn)));
    fmap.insert(make_pair(string("gnn"),    CFunc("gaussnn",  func_gaussnn)));
    fmap.insert(make_pair(string("cauchy"),CFunc("cauchy",  func_cauchy)));

    fmap.insert(make_pair(string("ran"),  CFunc("ran",  func_ran)));

    fmap.insert(make_pair(string("q4w"),  CFunc("ran",  func_q4w)));
    fmap.insert(make_pair(string("cauchy2"),  CFunc("cauchy2",  func_cauchy2)));

// f (data)
    fmap.insert(make_pair(string("intp"), CFunc("intp",func_data_intp_lin)));
}

CFunc* NFunctions::find(string s)
{
    // printf( "find\n" );
    map<string,CFunc>::iterator pos;
    if ((pos = fmap.find(s)) == (fmap.end())) return 0;
    // printf( "/find\n" );
    return &(pos->second);
}


//**************************************************************************//
//*   Coordinate concatenation                                             *//
//**************************************************************************//

CCoord CFunc::coord(CCoord *co1) const
{
    CCoord co = *co1;
    if        (s=="1/") {
        co.name = "1/" + co1->name;
        co.unit = co1->unit=="" ? "" : "(" + co1->unit + ")^-1";	
    } else if (s=="0-") {
        co.name = "-" + co1->name;
    } else if (s=="e^") {
        co.name = "e^" + co1->name;
    } else if (s=="10^") {
        co.name = "10^" + co1->name;
    } else if (s=="ln") {
        co.name = "ln " + co1->name;
    } else if (s=="lg") {
        co.name = "lg " + co1->name;
    } else if (s=="^2") {
        co.name = "(" + co1->name + ")^2";
        co.unit = co1->unit=="" ? "" : "(" + co1->unit + ")^2";	
    } else if (s=="sqrt") {
        co.name = "sqrt (" + co1->name + ")";
        co.unit = co1->unit=="" ? "" : "(" + co1->unit + ")^1/2";	
    } else if (s=="abs") {
        co.name = "abs (" + co1->name + ")";
    } else if (s=="") {
    }
    return co;
}

CCoord CFunc::coord(class CCoord *co1, class CCoord *co2) const
{
    CCoord co = *co1;

    if (s=="+" || s=="-") {
        co.name = co1->name + s + co2->name;
        if (co2->unit != co1->unit) {
            cout << "? different units " << co1->unit << " and " <<
                co2->unit << " in additive operation\n";
            co.unit = "";
        } else
            co.unit = co1->unit;
    } else if (s=="*" || s=="/") {
        co.name = co1->name + s + co2->name;
        if     ( co1->unit!="" && co2->unit!="" )
            co.unit = co1->unit + s + co2->unit;
        else if( co1->unit!="" )
            co.unit = co1->unit;
        else if( co2->unit!="" )
            co.unit = s=="*" ? co2->unit : "1/(" + co2->unit + ")";
        else
            co.unit = "";
    } else if (s=="^") {
        if (co2->unit != "") {
            cout << "? exponent is not dimensionless\n";
        }
        co.name = co1->name + "^" + co2->name;
        co.unit = co1->unit=="" ? "" : ( co1->unit + "^<val>" );
    }
    return co;
}

CCoord CFunc::coord(class CCoord *co1, class CCoord *co2,
                    class CCoord *co3) const
{
    CCoord co = *co2;
    return co;
}
