//**************************************************************************//
//* FRIDA: fast reliable interactive data analysis                         *//
//* list.h: the CList data type                                            *//
//* (C) Joachim Wuttke 1990-, v2(C++) 2001-                                *//
//* http://frida.sourceforge.net                                           *//
//**************************************************************************//

// A CList is basically a sequence of non-zero integers like 1-3,5,7,11.

// User-supplied input strings, however, may be more complicated:
// - input must not be ordered;
// - duplication of items is currently tolerated
//     (** this is NOT a design feature; it may change in the future **);
// - an interval can be qualified by a step: '1-100.2' means '1,3,5,...,99';
// - the last interval may read '101-*': it extends up to sup;
// - the user selection may just read '*': it extends from inf to sup;
// where inf and sup are context-dependent lower and upper bound.

// User-supplied bounds play two roles:
// - they define, how '-*' and '*' are interpreted;
// - in Ask(), they determine what answers are acceptable.
// These bounds can be set
// - by a constructor,
// - by Ask(),
// - by evaluate.
// They are valid until they are overwritten.
// To reset a CList into a bound-free state, it must be re-constructed.

class CList {
private:
    string instruction; // either "", or a user-supplied input string
    uint inf, sup;      // user-supplied bounds; meaningful only if 'limgiven'
    bool limgiven;      // have bounds been supplied ?
    bool limneeded;     // does instruction require bounds ?

public:
    vector<uint> V;     // holds the list, except if not yet 'evaluated()'

    CList();

    // initialize by string:
    CList( string inp );
    CList( string inp, uint _inf, uint _sup );

    // initialize by values:
    CList( uint val );
    CList( uint from, uint to );

    // set bounds and evaluate instruction:
    void evaluate( uint _inf, uint _sup );

    // re-arrange V (normally, the user-supplied order is retained)
    void Sort(void);
    void SortReverse(void);

    // append other sequence or item, unless duplicate:
    void append( const CList L );
    void append( const uint val );

    // inspect:
    inline bool evaluated() const { return !limneeded || limgiven; }
    inline uint size() const { return V.size(); };
    bool contained(uint val) const;
    bool all_between( const int imi, const int ima ); // OBSOLETE
    string str(void) const;
    inline friend ostream& operator<< (ostream& s, CList L) {
        return s << L.str(); };

private:
    void Parse(string inp);
    void ParseSkipinterval(string inp);
    int  ParseSkipfactor(string inp);
    void ParseInterval(string inp, int skip);
    int  ParseListvalue(string inp, int which);

    string sel2text(void) const;
};	
inline string strg(const CList& L) { return L.str(); }; // needed by myask

class ListIter { 
// OBSOLETE. will be eliminated as soon as ScanBreak is revised
private:
    uint i;
    CList L;
public:
    inline ListIter(void) { Reset(); };
    inline ListIter(CList L_in) { Reset(L_in); };
    inline void Reset(void) { i=0; };
    inline void Reset(CList L_in) { L = L_in; Reset(); };
    inline uint operator()(uint *val) {
        if (i>=L.size()) return 0;
        *val = L.V[i++];
        return i;
    };
};

