//**************************************************************************//
//* FRIDA: fast reliable interactive data analysis                         *//
//* xax.y: parser for eXtended Array eXpressions                           *//
//* (C) Joachim Wuttke 1990-, v2(C++) 2001-                                *//
//* http://frida.sourceforge.net                                           *//
//**************************************************************************//

%{
using namespace std;
#include <boost/shared_ptr.hpp>
#include <math.h>
#include <string.h>
#include <stdlib.h>
#include <string>
#include <vector> // for expr.h
#include "coord.h"
#include "func.h"
#include "expr.h"

// xaxtype must also be declared in xax.l
struct xaxtype {
    double  v;
    void*   p;
    PTree   t;
    PRef    r;
};
#define YYSTYPE xaxtype

#define YYERROR_VERBOSE 1
#define YYPARSE_PARAM T       /* used via xaxparse( CTree *T ) */
int xaxerror(const char*);    /* provided by calling code (expr.cpp, ...) */
int xaxlex(YYSTYPE *xaxlval); /* created by lex.l */

#define MONFUNC(op,v1)       PTree( new CTree(NFunctions::find(op),v1.t ) )
#define BINFUNC(op,v1,v2)    PTree( new CTree(NFunctions::find(op),v1.t,v2.t ) )
#define TERFUNC(op,v1,v2,v3) PTree( new CTree(NFunctions::find(op),v1.t,v2.t,\
  v3.t) )
%}

%pure-parser
%debug

/* Operator precedence */

%left '?' ':'
%left BOP_OR
%left BOP_AND
%left BOP_XOR
%left BOP_EQ BOP_NE
%left BOP_GT BOP_GE BOP_LT BOP_LE
%left '-' '+'
%left '*' '/' '%' BOP_MOD BOP_DIV
%left BOP_OVER /* binomial coefficient */
%left NEG     /* arithmetic negation (unary minus) */
%right '!'    /* logical not */
%right '^'    /* exponentiation */

/* other tokens created by lex.l: */
%token NUM REF ARG FRF FNCT END CONV DIRAC

/* Grammar follows */

%%
input: 
        exp END            {  *((PTree*)T) = $1.t; return 0; }
      | error              { yyerrok; return -1; }
;

exp:                       /* double-valued expression */
        NUM                { $$.t = PTree( new CTree($1.v) ); }
      | DIRAC              { $$.t = PTree( new CTree( 'd' ) ); }
      | iva                { $$.t = PTree( new CTree($1.r) ); }
      | ARG                { $$.t = PTree( new CTree( 't' ) ); }
      | ifu '(' exp ')'    { $$.t = PTree( new CTree($1.r, $3.t) ); }
      | CONV '(' exp ',' exp ')' { $$.t = 
                             PTree( new CTree( 'c', $3.t, $5.t ) ); }
      | DIRAC '(' exp ')'  { $$.t = PTree( new CTree( 'd', $3.t ) ); }
/*    | VAR                { $$ = $1->value.var;              } */
/*    | VAR '=' exp        { $$ = $3; $1->value.var = $3;     } */
      | FNCT '(' exp ')'   { $$.t = PTree( new CTree((CFunc*) $1.p, $3.t) ); }
      | FNCT '(' exp ',' exp ')' { $$.t =
                             PTree( new CTree((CFunc*) $1.p, $3.t, $5.t ) ); }
      | FNCT '(' exp ',' exp ',' exp ')' { $$.t = 
                             PTree( new CTree((CFunc*) $1.p,$3.t,$5.t,$7.t) ); }
      | exp '+' exp        { $$.t = BINFUNC("+", $1, $3); }
      | exp '-' exp        { $$.t = BINFUNC("-", $1, $3); }
      | exp '*' exp        { $$.t = BINFUNC("*", $1, $3); }
      | exp '/' exp        { $$.t = BINFUNC("/", $1, $3); }
      | exp '%' exp        { $$.t = BINFUNC("mod", $1, $3); }
      | exp '^' exp        { $$.t = BINFUNC("^", $1, $3); }
      | '-' exp  %prec NEG { $$.t = MONFUNC("0-", $2); }
      | exp BOP_EQ exp     { $$.t = BINFUNC("==",  $1, $3); }
      | exp BOP_NE exp     { $$.t = BINFUNC("!=",  $1, $3); }
      | exp BOP_GT exp     { $$.t = BINFUNC(">",  $1, $3); }
      | exp BOP_GE exp     { $$.t = BINFUNC(">=",  $1, $3); }
      | exp BOP_LT exp     { $$.t = BINFUNC("<=",  $1, $3); }
      | exp BOP_LE exp     { $$.t = BINFUNC("<",  $1, $3); }
      | exp BOP_AND exp    { $$.t = BINFUNC("&&", $1, $3); }
      | exp BOP_OR exp     { $$.t = BINFUNC("||",  $1, $3); }
      | exp BOP_XOR exp    { $$.t = BINFUNC("xor", $1, $3); }
      | exp BOP_MOD exp    { $$.t = BINFUNC("mod", $1, $3); }
      | exp BOP_DIV exp    { $$.t = BINFUNC("idiv",$1, $3); }
      | exp BOP_OVER exp   { $$.t = BINFUNC("over",$1, $3); }
      | '!' exp            { $$.t = MONFUNC("!", $2); }
      | exp '?' exp ':' exp {$$.t = TERFUNC("?:", $1, $3, $5); }
      | '(' exp ')'        { $$.t = $2.t; } 
;

iva:                       /* indexed variable */
        REF               
       { $$.r = PRef( new CRef(string((const char*)$1.p)) ); }
      | REF '[' rex ']'
       { $$.r = PRef( new CRef(string((const char*)$1.p), $3.t) ); }
      | REF '[' rex ',' rex ']'   
       { $$.r = PRef( new CRef(string((const char*)$1.p), $3.t, $5.t) ); }
      | REF '[' rex ',' rex ',' rex ']'   
       { $$.r = PRef( new CRef(string((const char*)$1.p), $3.t,$5.t,$7.t) ); }
;

ifu:                       /* indexed function */
        FRF                     { $$.r = PRef( new CRef( "f" ) ); }
      | FRF '[' rex ']'         { $$.r = PRef( new CRef( "f", $3.t) ); }
      | FRF '[' rex ',' rex ']' { $$.r = PRef( new CRef( "f", $3.t, $5.t) ); }
;

rex:                       /* reference expression */
        /* empty */        { $$.r = PRef();}
      | exp                { $$.r = $1.r; }
;

/* End of grammar */
%%
